# -*- coding: utf-8 -*-
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.

import sys
import datetime

from transmissionrpc.constants import PRIORITY, RATIO_LIMIT, IDLE_LIMIT
from transmissionrpc.utils import Field, format_timedelta


def get_status_old(code):
    """Get the torrent status using old status codes"""
    mapping = {
        (1 << 0): "check pending",
        (1 << 1): "checking",
        (1 << 2): "downloading",
        (1 << 3): "seeding",
        (1 << 4): "stopped",
    }
    return mapping[code]


def get_status_new(code):
    """Get the torrent status using new status codes"""
    mapping = {
        0: "stopped",
        1: "check pending",
        2: "checking",
        3: "download pending",
        4: "downloading",
        5: "seed pending",
        6: "seeding",
    }
    return mapping[code]


class Torrent:
    """
    Torrent is a class holding the data received from Transmission regarding a bittorrent transfer.

    All fetched torrent fields are accessible through this class using attributes.
    This class has a few convenience properties using the torrent data.
    """

    def __init__(self, client, fields):
        if "id" not in fields:
            raise ValueError("Torrent requires an id")
        self._fields = {}
        self._update_fields(fields)
        self._incoming_pending = False
        self._outgoing_pending = False
        self._client = client

    def _get_name_string(self, codec=None):
        """Get the name"""
        if codec is None:
            codec = sys.getdefaultencoding()
        name = None
        # try to find name
        if "name" in self._fields:
            name = self._fields["name"].value
        # if name is unicode, try to decode
        if isinstance(name, str):
            try:
                name = name.encode(codec)
            except UnicodeError:
                name = None
        return name

    def __repr__(self):
        tid = self._fields["id"].value
        name = self._get_name_string()
        if isinstance(name, str):
            return '<Torrent %d "%s">' % (tid, name)
        else:
            return "<Torrent %d>" % (tid)

    def __str__(self):
        name = self._get_name_string()
        if isinstance(name, str):
            return 'Torrent "%s"' % (name)
        else:
            return "Torrent"

    def __copy__(self):
        return Torrent(self._client, self._fields)

    def __getattr__(self, name):
        try:
            return self._fields[name].value
        except KeyError:
            raise AttributeError("No attribute %s" % name)

    def _rpc_version(self):
        """Get the Transmission RPC API version."""
        if self._client:
            return self._client.rpc_version
        return 2

    def _dirty_fields(self):
        """Enumerate changed fields"""
        outgoing_keys = [
            "bandwidthPriority",
            "downloadLimit",
            "downloadLimited",
            "peer_limit",
            "queuePosition",
            "seedIdleLimit",
            "seedIdleMode",
            "seedRatioLimit",
            "seedRatioMode",
            "uploadLimit",
            "uploadLimited",
        ]
        fields = []
        for key in outgoing_keys:
            if key in self._fields and self._fields[key].dirty:
                fields.append(key)
        return fields

    def _push(self):
        """Push changed fields to the server"""
        dirty = self._dirty_fields()
        args = {}
        for key in dirty:
            args[key] = self._fields[key].value
            self._fields[key] = self._fields[key]._replace(dirty=False)
        if args:
            self._client.change_torrent(self.id, **args)

    def _update_fields(self, other):
        """
        Update the torrent data from a Transmission JSON-RPC arguments dictionary
        """
        if isinstance(other, dict):
            for key, value in list(other.items()):
                self._fields[key.replace("-", "_")] = Field(value, False)
        elif isinstance(other, Torrent):
            for key in list(other._fields.keys()):
                self._fields[key] = Field(other._fields[key].value, False)
        else:
            raise ValueError("Cannot update with supplied data")
        self._incoming_pending = False

    def _status(self):
        """Get the torrent status"""
        code = self._fields["status"].value
        if self._rpc_version() >= 14:
            return get_status_new(code)
        else:
            return get_status_old(code)

    def files(self):
        """
        Get list of files for this torrent.

        This function returns a dictionary with file information for each file.
        The file information is has following fields:
        ::

            {
                <file id>: {
                    'name': <file name>,
                    'size': <file size in bytes>,
                    'completed': <bytes completed>,
                    'priority': <priority ('high'|'normal'|'low')>,
                    'selected': <selected for download>
                }
                ...
            }
        """
        result = {}
        if "files" in self._fields:
            files = self._fields["files"].value
            indices = list(range(len(files)))
            priorities = self._fields["priorities"].value
            wanted = self._fields["wanted"].value
            for item in zip(indices, files, priorities, wanted):
                selected = True if item[3] else False
                priority = PRIORITY[item[2]]
                result[item[0]] = {
                    "selected": selected,
                    "priority": priority,
                    "size": item[1]["length"],
                    "name": item[1]["name"],
                    "completed": item[1]["bytesCompleted"],
                }
        return result

    @property
    def status(self):
        """
        Returns the torrent status. Is either one of 'check pending', 'checking',
        'downloading', 'seeding' or 'stopped'. The first two is related to
        verification.
        """
        return self._status()

    @property
    def progress(self):
        """Get the download progress in percent."""
        try:
            size = self._fields["sizeWhenDone"].value
            left = self._fields["leftUntilDone"].value
            return 100.0 * (size - left) / float(size)
        except ZeroDivisionError:
            return 0.0

    @property
    def ratio(self):
        """Get the upload/download ratio."""
        return float(self._fields["uploadRatio"].value)

    @property
    def eta(self):
        """Get the "eta" as datetime.timedelta."""
        eta = self._fields["eta"].value
        if eta >= 0:
            return datetime.timedelta(seconds=eta)
        else:
            raise ValueError("eta not valid")

    @property
    def date_active(self):
        """Get the attribute "activityDate" as datetime.datetime."""
        return datetime.datetime.fromtimestamp(self._fields["activityDate"].value)

    @property
    def date_added(self):
        """Get the attribute "addedDate" as datetime.datetime."""
        return datetime.datetime.fromtimestamp(self._fields["addedDate"].value)

    @property
    def date_started(self):
        """Get the attribute "startDate" as datetime.datetime."""
        return datetime.datetime.fromtimestamp(self._fields["startDate"].value)

    @property
    def date_done(self):
        """Get the attribute "doneDate" as datetime.datetime. returns None if "doneDate" is invalid."""
        done_date = self._fields["doneDate"].value
        # Transmission might forget to set doneDate which is initialized to zero, so if doneDate is zero return None
        if done_date == 0:
            return None
        else:
            return datetime.datetime.fromtimestamp(done_date)

    def format_eta(self):
        """
        Returns the attribute *eta* formatted as a string.

        * If eta is -1 the result is 'not available'
        * If eta is -2 the result is 'unknown'
        * Otherwise eta is formatted as <days> <hours>:<minutes>:<seconds>.
        """
        eta = self._fields["eta"].value
        if eta == -1:
            return "not available"
        elif eta == -2:
            return "unknown"
        else:
            return format_timedelta(self.eta)

    def _get_download_limit(self):
        """
        Get the download limit.
        Can be a number or None.
        """
        if self._fields["downloadLimited"].value:
            return self._fields["downloadLimit"].value
        else:
            return None

    def _set_download_limit(self, limit):
        """
        Get the download limit.
        Can be a number, 'session' or None.
        """
        if isinstance(limit, int):
            self._fields["downloadLimited"] = Field(True, True)
            self._fields["downloadLimit"] = Field(limit, True)
            self._push()
        elif limit is None:
            self._fields["downloadLimited"] = Field(False, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    download_limit = property(
        _get_download_limit,
        _set_download_limit,
        None,
        "Download limit in Kbps or None. This is a mutator.",
    )

    def _get_peer_limit(self):
        """
        Get the peer limit.
        """
        return self._fields["peer_limit"].value

    def _set_peer_limit(self, limit):
        """
        Set the peer limit.
        """
        if isinstance(limit, int):
            self._fields["peer_limit"] = Field(limit, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    peer_limit = property(
        _get_peer_limit, _set_peer_limit, None, "Peer limit. This is a mutator."
    )

    def _get_priority(self):
        """
        Get the priority as string.
        Can be one of 'low', 'normal', 'high'.
        """
        return PRIORITY[self._fields["bandwidthPriority"].value]

    def _set_priority(self, priority):
        """
        Set the priority as string.
        Can be one of 'low', 'normal', 'high'.
        """
        if isinstance(priority, str):
            self._fields["bandwidthPriority"] = Field(PRIORITY[priority], True)
            self._push()

    priority = property(
        _get_priority,
        _set_priority,
        None,
        "Bandwidth priority as string. Can be one of 'low', 'normal', 'high'. This is a mutator.",
    )

    def _get_seed_idle_limit(self):
        """
        Get the seed idle limit in minutes.
        """
        return self._fields["seedIdleLimit"].value

    def _set_seed_idle_limit(self, limit):
        """
        Set the seed idle limit in minutes.
        """
        if isinstance(limit, int):
            self._fields["seedIdleLimit"] = Field(limit, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    seed_idle_limit = property(
        _get_seed_idle_limit,
        _set_seed_idle_limit,
        None,
        "Torrent seed idle limit in minutes. Also see seed_idle_mode. This is a mutator.",
    )

    def _get_seed_idle_mode(self):
        """
        Get the seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.
        """
        return IDLE_LIMIT[self._fields["seedIdleMode"].value]

    def _set_seed_idle_mode(self, mode):
        """
        Set the seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.
        """
        if isinstance(mode, str):
            self._fields["seedIdleMode"] = Field(IDLE_LIMIT[mode], True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    seed_idle_mode = property(
        _get_seed_idle_mode,
        _set_seed_idle_mode,
        None,
        """
        Seed idle mode as string. Can be one of 'global', 'single' or 'unlimited'.

         * global, use session seed idle limit.
         * single, use torrent seed idle limit. See seed_idle_limit.
         * unlimited, no seed idle limit.

        This is a mutator.
        """,
    )

    def _get_seed_ratio_limit(self):
        """
        Get the seed ratio limit as float.
        """
        return float(self._fields["seedRatioLimit"].value)

    def _set_seed_ratio_limit(self, limit):
        """
        Set the seed ratio limit as float.
        """
        if isinstance(limit, (int, float)) and limit >= 0.0:
            self._fields["seedRatioLimit"] = Field(float(limit), True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    seed_ratio_limit = property(
        _get_seed_ratio_limit,
        _set_seed_ratio_limit,
        None,
        "Torrent seed ratio limit as float. Also see seed_ratio_mode. This is a mutator.",
    )

    def _get_seed_ratio_mode(self):
        """
        Get the seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.
        """
        return RATIO_LIMIT[self._fields["seedRatioMode"].value]

    def _set_seed_ratio_mode(self, mode):
        """
        Set the seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.
        """
        if isinstance(mode, str):
            self._fields["seedRatioMode"] = Field(RATIO_LIMIT[mode], True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    seed_ratio_mode = property(
        _get_seed_ratio_mode,
        _set_seed_ratio_mode,
        None,
        """
        Seed ratio mode as string. Can be one of 'global', 'single' or 'unlimited'.

         * global, use session seed ratio limit.
         * single, use torrent seed ratio limit. See seed_ratio_limit.
         * unlimited, no seed ratio limit.

        This is a mutator.
        """,
    )

    def _get_upload_limit(self):
        """
        Get the upload limit.
        Can be a number or None.
        """
        if self._fields["uploadLimited"].value:
            return self._fields["uploadLimit"].value
        else:
            return None

    def _set_upload_limit(self, limit):
        """
        Set the upload limit.
        Can be a number, 'session' or None.
        """
        if isinstance(limit, int):
            self._fields["uploadLimited"] = Field(True, True)
            self._fields["uploadLimit"] = Field(limit, True)
            self._push()
        elif limit is None:
            self._fields["uploadLimited"] = Field(False, True)
            self._push()
        else:
            raise ValueError("Not a valid limit")

    upload_limit = property(
        _get_upload_limit,
        _set_upload_limit,
        None,
        "Upload limit in Kbps or None. This is a mutator.",
    )

    def _get_queue_position(self):
        """Get the queue position for this torrent."""
        if self._rpc_version() >= 14:
            return self._fields["queuePosition"].value
        else:
            return 0

    def _set_queue_position(self, position):
        """Set the queue position for this torrent."""
        if self._rpc_version() >= 14:
            if isinstance(position, int):
                self._fields["queuePosition"] = Field(position, True)
                self._push()
            else:
                raise ValueError("Not a valid position")
        else:
            pass

    queue_position = property(
        _get_queue_position, _set_queue_position, None, "Queue position"
    )

    def update(self, timeout=None):
        """Update the torrent information."""
        self._push()
        torrent = self._client.get_torrent(self.id, timeout=timeout)
        self._update_fields(torrent)

    def start(self, bypass_queue=False, timeout=None):
        """
        Start the torrent.
        """
        self._incoming_pending = True
        self._client.start_torrent(self.id, bypass_queue=bypass_queue, timeout=timeout)

    def stop(self, timeout=None):
        """Stop the torrent."""
        self._incoming_pending = True
        self._client.stop_torrent(self.id, timeout=timeout)

    def move_data(self, location, timeout=None):
        """Move torrent data to location."""
        self._incoming_pending = True
        self._client.move_torrent_data(self.id, location, timeout=timeout)

    def locate_data(self, location, timeout=None):
        """Locate torrent data at location."""
        self._incoming_pending = True
        self._client.locate_torrent_data(self.id, location, timeout=timeout)



# -*- coding: utf-8 -*-
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.
# pylint: disable=line-too-long

import logging

LOGGER = logging.getLogger("transmissionrpc")
LOGGER.setLevel(logging.ERROR)


def mirror_dict(source):
    """
    Creates a dictionary with all values as keys and all keys as values.
    """
    source.update(dict((value, key) for key, value in list(source.items())))
    return source


DEFAULT_PORT = 9091

DEFAULT_TIMEOUT = 30.0

TR_PRI_LOW = -1
TR_PRI_NORMAL = 0
TR_PRI_HIGH = 1

PRIORITY = mirror_dict(
    {"low": TR_PRI_LOW, "normal": TR_PRI_NORMAL, "high": TR_PRI_HIGH}
)

TR_RATIOLIMIT_GLOBAL = 0  # follow the global settings
TR_RATIOLIMIT_SINGLE = 1  # override the global settings, seeding until a certain ratio
TR_RATIOLIMIT_UNLIMITED = 2  # override the global settings, seeding regardless of ratio

RATIO_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)

TR_IDLELIMIT_GLOBAL = 0  # follow the global settings
TR_IDLELIMIT_SINGLE = (
    1  # override the global settings, seeding until a certain idle time
)
TR_IDLELIMIT_UNLIMITED = (
    2  # override the global settings, seeding regardless of activity
)

IDLE_LIMIT = mirror_dict(
    {
        "global": TR_RATIOLIMIT_GLOBAL,
        "single": TR_RATIOLIMIT_SINGLE,
        "unlimited": TR_RATIOLIMIT_UNLIMITED,
    }
)

# A note on argument maps
# These maps are used to verify *-set methods. The information is structured in
# a tree.
# set +- <argument1> - [<type>, <added version>, <removed version>, <previous argument name>, <next argument name>, <description>]
#  |  +- <argument2> - [<type>, <added version>, <removed version>, <previous argument name>, <next argument name>, <description>]
#  |
# get +- <argument1> - [<type>, <added version>, <removed version>, <previous argument name>, <next argument name>, <description>]
#     +- <argument2> - [<type>, <added version>, <removed version>, <previous argument name>, <next argument name>, <description>]

# Arguments for torrent methods
TORRENT_ARGS = {
    "get": {
        "activityDate": (
            "number",
            1,
            None,
            None,
            None,
            "Last time of upload or download activity.",
        ),
        "addedDate": (
            "number",
            1,
            None,
            None,
            None,
            "The date when this torrent was first added.",
        ),
        "announceResponse": (
            "string",
            1,
            7,
            None,
            None,
            "The announce message from the tracker.",
        ),
        "announceURL": ("string", 1, 7, None, None, "Current announce URL."),
        "bandwidthPriority": (
            "number",
            5,
            None,
            None,
            None,
            "Bandwidth priority. Low (-1), Normal (0) or High (1).",
        ),
        "comment": ("string", 1, None, None, None, "Torrent comment."),
        "corruptEver": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes of corrupt data downloaded.",
        ),
        "creator": ("string", 1, None, None, None, "Torrent creator."),
        "dateCreated": ("number", 1, None, None, None, "Torrent creation date."),
        "desiredAvailable": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes avalable and left to be downloaded.",
        ),
        "doneDate": (
            "number",
            1,
            None,
            None,
            None,
            "The date when the torrent finished downloading.",
        ),
        "downloadDir": (
            "string",
            4,
            None,
            None,
            None,
            "The directory path where the torrent is downloaded to.",
        ),
        "downloadedEver": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes of good data downloaded.",
        ),
        "downloaders": ("number", 4, 7, None, None, "Number of downloaders."),
        "downloadLimit": ("number", 1, None, None, None, "Download limit in Kbps."),
        "downloadLimited": (
            "boolean",
            5,
            None,
            None,
            None,
            "Download limit is enabled",
        ),
        "downloadLimitMode": (
            "number",
            1,
            5,
            None,
            None,
            "Download limit mode. 0 means global, 1 means signle, 2 unlimited.",
        ),
        "error": (
            "number",
            1,
            None,
            None,
            None,
            "Kind of error. 0 means OK, 1 means tracker warning, 2 means tracker error, 3 means local error.",
        ),
        "errorString": ("number", 1, None, None, None, "Error message."),
        "eta": (
            "number",
            1,
            None,
            None,
            None,
            "Estimated number of seconds left when downloading or seeding. -1 means not available and -2 means unknown.",
        ),
        "etaIdle": (
            "number",
            15,
            None,
            None,
            None,
            "Estimated number of seconds left until the idle time limit is reached. -1 means not available and -2 means unknown.",
        ),
        "files": (
            "array",
            1,
            None,
            None,
            None,
            "Array of file object containing key, bytesCompleted, length and name.",
        ),
        "fileStats": (
            "array",
            5,
            None,
            None,
            None,
            "Aray of file statistics containing bytesCompleted, wanted and priority.",
        ),
        "hashString": (
            "string",
            1,
            None,
            None,
            None,
            "Hashstring unique for the torrent even between sessions.",
        ),
        "haveUnchecked": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes of partial pieces.",
        ),
        "haveValid": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes of checksum verified data.",
        ),
        "honorsSessionLimits": (
            "boolean",
            5,
            None,
            None,
            None,
            "True if session upload limits are honored",
        ),
        "id": ("number", 1, None, None, None, "Session unique torrent id."),
        "isFinished": (
            "boolean",
            9,
            None,
            None,
            None,
            "True if the torrent is finished. Downloaded and seeded.",
        ),
        "isPrivate": (
            "boolean",
            1,
            None,
            None,
            None,
            "True if the torrent is private.",
        ),
        "isStalled": (
            "boolean",
            14,
            None,
            None,
            None,
            "True if the torrent has stalled (been idle for a long time).",
        ),
        "lastAnnounceTime": (
            "number",
            1,
            7,
            None,
            None,
            "The time of the last announcement.",
        ),
        "lastScrapeTime": (
            "number",
            1,
            7,
            None,
            None,
            "The time af the last successful scrape.",
        ),
        "leechers": ("number", 1, 7, None, None, "Number of leechers."),
        "leftUntilDone": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes left until the download is done.",
        ),
        "magnetLink": (
            "string",
            7,
            None,
            None,
            None,
            "The magnet link for this torrent.",
        ),
        "manualAnnounceTime": (
            "number",
            1,
            None,
            None,
            None,
            "The time until you manually ask for more peers.",
        ),
        "maxConnectedPeers": (
            "number",
            1,
            None,
            None,
            None,
            "Maximum of connected peers.",
        ),
        "metadataPercentComplete": (
            "number",
            7,
            None,
            None,
            None,
            "Download progress of metadata. 0.0 to 1.0.",
        ),
        "name": ("string", 1, None, None, None, "Torrent name."),
        "nextAnnounceTime": ("number", 1, 7, None, None, "Next announce time."),
        "nextScrapeTime": ("number", 1, 7, None, None, "Next scrape time."),
        "peer-limit": ("number", 5, None, None, None, "Maximum number of peers."),
        "peers": ("array", 2, None, None, None, "Array of peer objects."),
        "peersConnected": (
            "number",
            1,
            None,
            None,
            None,
            "Number of peers we are connected to.",
        ),
        "peersFrom": (
            "object",
            1,
            None,
            None,
            None,
            "Object containing download peers counts for different peer types.",
        ),
        "peersGettingFromUs": (
            "number",
            1,
            None,
            None,
            None,
            "Number of peers we are sending data to.",
        ),
        "peersKnown": (
            "number",
            1,
            13,
            None,
            None,
            "Number of peers that the tracker knows.",
        ),
        "peersSendingToUs": (
            "number",
            1,
            None,
            None,
            None,
            "Number of peers sending to us",
        ),
        "percentDone": (
            "double",
            5,
            None,
            None,
            None,
            "Download progress of selected files. 0.0 to 1.0.",
        ),
        "pieces": (
            "string",
            5,
            None,
            None,
            None,
            "String with base64 encoded bitfield indicating finished pieces.",
        ),
        "pieceCount": ("number", 1, None, None, None, "Number of pieces."),
        "pieceSize": ("number", 1, None, None, None, "Number of bytes in a piece."),
        "priorities": ("array", 1, None, None, None, "Array of file priorities."),
        "queuePosition": ("number", 14, None, None, None, "The queue position."),
        "rateDownload": ("number", 1, None, None, None, "Download rate in bps."),
        "rateUpload": ("number", 1, None, None, None, "Upload rate in bps."),
        "recheckProgress": (
            "double",
            1,
            None,
            None,
            None,
            "Progress of recheck. 0.0 to 1.0.",
        ),
        "secondsDownloading": ("number", 15, None, None, None, ""),
        "secondsSeeding": ("number", 15, None, None, None, ""),
        "scrapeResponse": ("string", 1, 7, None, None, "Scrape response message."),
        "scrapeURL": ("string", 1, 7, None, None, "Current scrape URL"),
        "seeders": (
            "number",
            1,
            7,
            None,
            None,
            "Number of seeders reported by the tracker.",
        ),
        "seedIdleLimit": ("number", 10, None, None, None, "Idle limit in minutes."),
        "seedIdleMode": (
            "number",
            10,
            None,
            None,
            None,
            "Use global (0), torrent (1), or unlimited (2) limit.",
        ),
        "seedRatioLimit": ("double", 5, None, None, None, "Seed ratio limit."),
        "seedRatioMode": (
            "number",
            5,
            None,
            None,
            None,
            "Use global (0), torrent (1), or unlimited (2) limit.",
        ),
        "sizeWhenDone": (
            "number",
            1,
            None,
            None,
            None,
            "Size of the torrent download in bytes.",
        ),
        "startDate": (
            "number",
            1,
            None,
            None,
            None,
            "The date when the torrent was last started.",
        ),
        "status": ("number", 1, None, None, None, "Current status, see source"),
        "swarmSpeed": (
            "number",
            1,
            7,
            None,
            None,
            "Estimated speed in Kbps in the swarm.",
        ),
        "timesCompleted": (
            "number",
            1,
            7,
            None,
            None,
            "Number of successful downloads reported by the tracker.",
        ),
        "trackers": ("array", 1, None, None, None, "Array of tracker objects."),
        "trackerStats": (
            "object",
            7,
            None,
            None,
            None,
            "Array of object containing tracker statistics.",
        ),
        "totalSize": (
            "number",
            1,
            None,
            None,
            None,
            "Total size of the torrent in bytes",
        ),
        "torrentFile": ("string", 5, None, None, None, "Path to .torrent file."),
        "uploadedEver": (
            "number",
            1,
            None,
            None,
            None,
            "Number of bytes uploaded, ever.",
        ),
        "uploadLimit": ("number", 1, None, None, None, "Upload limit in Kbps"),
        "uploadLimitMode": (
            "number",
            1,
            5,
            None,
            None,
            "Upload limit mode. 0 means global, 1 means signle, 2 unlimited.",
        ),
        "uploadLimited": ("boolean", 5, None, None, None, "Upload limit enabled."),
        "uploadRatio": ("double", 1, None, None, None, "Seed ratio."),
        "wanted": (
            "array",
            1,
            None,
            None,
            None,
            "Array of booleans indicated wanted files.",
        ),
        "webseeds": ("array", 1, None, None, None, "Array of webseeds objects"),
        "webseedsSendingToUs": (
            "number",
            1,
            None,
            None,
            None,
            "Number of webseeds seeding to us.",
        ),
    },
    "set": {
        "bandwidthPriority": (
            "number",
            5,
            None,
            None,
            None,
            "Priority for this transfer.",
        ),
        "downloadLimit": (
            "number",
            5,
            None,
            "speed-limit-down",
            None,
            "Set the speed limit for download in Kib/s.",
        ),
        "downloadLimited": (
            "boolean",
            5,
            None,
            "speed-limit-down-enabled",
            None,
            "Enable download speed limiter.",
        ),
        "files-wanted": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should be downloaded.",
        ),
        "files-unwanted": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that shouldn't be downloaded.",
        ),
        "honorsSessionLimits": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables or disables the transfer to honour the upload limit set in the session.",
        ),
        "location": ("array", 1, None, None, None, "Local download location."),
        "peer-limit": (
            "number",
            1,
            None,
            None,
            None,
            "The peer limit for the torrents.",
        ),
        "priority-high": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have high priority.",
        ),
        "priority-low": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have normal priority.",
        ),
        "priority-normal": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have low priority.",
        ),
        "queuePosition": (
            "number",
            14,
            None,
            None,
            None,
            "Position of this transfer in its queue.",
        ),
        "seedIdleLimit": (
            "number",
            10,
            None,
            None,
            None,
            "Seed inactivity limit in minutes.",
        ),
        "seedIdleMode": (
            "number",
            10,
            None,
            None,
            None,
            "Seed inactivity mode. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.",
        ),
        "seedRatioLimit": ("double", 5, None, None, None, "Seeding ratio."),
        "seedRatioMode": (
            "number",
            5,
            None,
            None,
            None,
            "Which ratio to use. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.",
        ),
        "speed-limit-down": (
            "number",
            1,
            5,
            None,
            "downloadLimit",
            "Set the speed limit for download in Kib/s.",
        ),
        "speed-limit-down-enabled": (
            "boolean",
            1,
            5,
            None,
            "downloadLimited",
            "Enable download speed limiter.",
        ),
        "speed-limit-up": (
            "number",
            1,
            5,
            None,
            "uploadLimit",
            "Set the speed limit for upload in Kib/s.",
        ),
        "speed-limit-up-enabled": (
            "boolean",
            1,
            5,
            None,
            "uploadLimited",
            "Enable upload speed limiter.",
        ),
        "trackerAdd": (
            "array",
            10,
            None,
            None,
            None,
            "Array of string with announce URLs to add.",
        ),
        "trackerRemove": (
            "array",
            10,
            None,
            None,
            None,
            "Array of ids of trackers to remove.",
        ),
        "trackerReplace": (
            "array",
            10,
            None,
            None,
            None,
            "Array of (id, url) tuples where the announce URL should be replaced.",
        ),
        "uploadLimit": (
            "number",
            5,
            None,
            "speed-limit-up",
            None,
            "Set the speed limit for upload in Kib/s.",
        ),
        "uploadLimited": (
            "boolean",
            5,
            None,
            "speed-limit-up-enabled",
            None,
            "Enable upload speed limiter.",
        ),
    },
    "add": {
        "bandwidthPriority": (
            "number",
            8,
            None,
            None,
            None,
            "Priority for this transfer.",
        ),
        "download-dir": (
            "string",
            1,
            None,
            None,
            None,
            "The directory where the downloaded contents will be saved in.",
        ),
        "cookies": ("string", 13, None, None, None, "One or more HTTP cookie(s)."),
        "filename": (
            "string",
            1,
            None,
            None,
            None,
            "A file path or URL to a torrent file or a magnet link.",
        ),
        "files-wanted": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should be downloaded.",
        ),
        "files-unwanted": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that shouldn't be downloaded.",
        ),
        "metainfo": (
            "string",
            1,
            None,
            None,
            None,
            "The content of a torrent file, base64 encoded.",
        ),
        "paused": (
            "boolean",
            1,
            None,
            None,
            None,
            "If True, does not start the transfer when added.",
        ),
        "peer-limit": (
            "number",
            1,
            None,
            None,
            None,
            "Maximum number of peers allowed.",
        ),
        "priority-high": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have high priority.",
        ),
        "priority-low": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have low priority.",
        ),
        "priority-normal": (
            "array",
            1,
            None,
            None,
            None,
            "A list of file id's that should have normal priority.",
        ),
    },
}

# Arguments for session methods
SESSION_ARGS = {
    "get": {
        "alt-speed-down": (
            "number",
            5,
            None,
            None,
            None,
            "Alternate session download speed limit (in Kib/s).",
        ),
        "alt-speed-enabled": (
            "boolean",
            5,
            None,
            None,
            None,
            "True if alternate global download speed limiter is ebabled.",
        ),
        "alt-speed-time-begin": (
            "number",
            5,
            None,
            None,
            None,
            "Time when alternate speeds should be enabled. Minutes after midnight.",
        ),
        "alt-speed-time-enabled": (
            "boolean",
            5,
            None,
            None,
            None,
            "True if alternate speeds scheduling is enabled.",
        ),
        "alt-speed-time-end": (
            "number",
            5,
            None,
            None,
            None,
            "Time when alternate speeds should be disabled. Minutes after midnight.",
        ),
        "alt-speed-time-day": (
            "number",
            5,
            None,
            None,
            None,
            "Days alternate speeds scheduling is enabled.",
        ),
        "alt-speed-up": (
            "number",
            5,
            None,
            None,
            None,
            "Alternate session upload speed limit (in Kib/s)",
        ),
        "blocklist-enabled": (
            "boolean",
            5,
            None,
            None,
            None,
            "True when blocklist is enabled.",
        ),
        "blocklist-size": (
            "number",
            5,
            None,
            None,
            None,
            "Number of rules in the blocklist",
        ),
        "blocklist-url": (
            "string",
            11,
            None,
            None,
            None,
            "Location of the block list. Updated with blocklist-update.",
        ),
        "cache-size-mb": (
            "number",
            10,
            None,
            None,
            None,
            "The maximum size of the disk cache in MB",
        ),
        "config-dir": (
            "string",
            8,
            None,
            None,
            None,
            "location of transmissions configuration directory",
        ),
        "dht-enabled": ("boolean", 6, None, None, None, "True if DHT enabled."),
        "download-dir": ("string", 1, None, None, None, "The download directory."),
        "download-dir-free-space": (
            "number",
            12,
            None,
            None,
            None,
            "Free space in the download directory, in bytes",
        ),
        "download-queue-size": (
            "number",
            14,
            None,
            None,
            None,
            "Number of slots in the download queue.",
        ),
        "download-queue-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "True if the download queue is enabled.",
        ),
        "encryption": (
            "string",
            1,
            None,
            None,
            None,
            "Encryption mode, one of ``required``, ``preferred`` or ``tolerated``.",
        ),
        "idle-seeding-limit": (
            "number",
            10,
            None,
            None,
            None,
            "Seed inactivity limit in minutes.",
        ),
        "idle-seeding-limit-enabled": (
            "boolean",
            10,
            None,
            None,
            None,
            "True if the seed activity limit is enabled.",
        ),
        "incomplete-dir": (
            "string",
            7,
            None,
            None,
            None,
            "The path to the directory for incomplete torrent transfer data.",
        ),
        "incomplete-dir-enabled": (
            "boolean",
            7,
            None,
            None,
            None,
            "True if the incomplete dir is enabled.",
        ),
        "lpd-enabled": (
            "boolean",
            9,
            None,
            None,
            None,
            "True if local peer discovery is enabled.",
        ),
        "peer-limit": (
            "number",
            1,
            5,
            None,
            "peer-limit-global",
            "Maximum number of peers.",
        ),
        "peer-limit-global": (
            "number",
            5,
            None,
            "peer-limit",
            None,
            "Maximum number of peers.",
        ),
        "peer-limit-per-torrent": (
            "number",
            5,
            None,
            None,
            None,
            "Maximum number of peers per transfer.",
        ),
        "pex-allowed": (
            "boolean",
            1,
            5,
            None,
            "pex-enabled",
            "True if PEX is allowed.",
        ),
        "pex-enabled": (
            "boolean",
            5,
            None,
            "pex-allowed",
            None,
            "True if PEX is enabled.",
        ),
        "port": ("number", 1, 5, None, "peer-port", "Peer port."),
        "peer-port": ("number", 5, None, "port", None, "Peer port."),
        "peer-port-random-on-start": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables randomized peer port on start of Transmission.",
        ),
        "port-forwarding-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "True if port forwarding is enabled.",
        ),
        "queue-stalled-minutes": (
            "number",
            14,
            None,
            None,
            None,
            "Number of minutes of idle that marks a transfer as stalled.",
        ),
        "queue-stalled-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "True if stalled tracking of transfers is enabled.",
        ),
        "rename-partial-files": (
            "boolean",
            8,
            None,
            None,
            None,
            'True if ".part" is appended to incomplete files',
        ),
        "rpc-version": ("number", 4, None, None, None, "Transmission RPC API Version."),
        "rpc-version-minimum": (
            "number",
            4,
            None,
            None,
            None,
            "Minimum accepted RPC API Version.",
        ),
        "script-torrent-done-enabled": (
            "boolean",
            9,
            None,
            None,
            None,
            "True if the done script is enabled.",
        ),
        "script-torrent-done-filename": (
            "string",
            9,
            None,
            None,
            None,
            "Filename of the script to run when the transfer is done.",
        ),
        "seedRatioLimit": (
            "double",
            5,
            None,
            None,
            None,
            "Seed ratio limit. 1.0 means 1:1 download and upload ratio.",
        ),
        "seedRatioLimited": (
            "boolean",
            5,
            None,
            None,
            None,
            "True if seed ration limit is enabled.",
        ),
        "seed-queue-size": (
            "number",
            14,
            None,
            None,
            None,
            "Number of slots in the upload queue.",
        ),
        "seed-queue-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "True if upload queue is enabled.",
        ),
        "speed-limit-down": (
            "number",
            1,
            None,
            None,
            None,
            "Download speed limit (in Kib/s).",
        ),
        "speed-limit-down-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "True if the download speed is limited.",
        ),
        "speed-limit-up": (
            "number",
            1,
            None,
            None,
            None,
            "Upload speed limit (in Kib/s).",
        ),
        "speed-limit-up-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "True if the upload speed is limited.",
        ),
        "start-added-torrents": (
            "boolean",
            9,
            None,
            None,
            None,
            "When true uploaded torrents will start right away.",
        ),
        "trash-original-torrent-files": (
            "boolean",
            9,
            None,
            None,
            None,
            "When true added .torrent files will be deleted.",
        ),
        "units": (
            "object",
            10,
            None,
            None,
            None,
            "An object containing units for size and speed.",
        ),
        "utp-enabled": (
            "boolean",
            13,
            None,
            None,
            None,
            "True if Micro Transport Protocol (UTP) is enabled.",
        ),
        "version": ("string", 3, None, None, None, "Transmission version."),
    },
    "set": {
        "alt-speed-down": (
            "number",
            5,
            None,
            None,
            None,
            "Alternate session download speed limit (in Kib/s).",
        ),
        "alt-speed-enabled": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables alternate global download speed limiter.",
        ),
        "alt-speed-time-begin": (
            "number",
            5,
            None,
            None,
            None,
            "Time when alternate speeds should be enabled. Minutes after midnight.",
        ),
        "alt-speed-time-enabled": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables alternate speeds scheduling.",
        ),
        "alt-speed-time-end": (
            "number",
            5,
            None,
            None,
            None,
            "Time when alternate speeds should be disabled. Minutes after midnight.",
        ),
        "alt-speed-time-day": (
            "number",
            5,
            None,
            None,
            None,
            "Enables alternate speeds scheduling these days.",
        ),
        "alt-speed-up": (
            "number",
            5,
            None,
            None,
            None,
            "Alternate session upload speed limit (in Kib/s).",
        ),
        "blocklist-enabled": ("boolean", 5, None, None, None, "Enables the block list"),
        "blocklist-url": (
            "string",
            11,
            None,
            None,
            None,
            "Location of the block list. Updated with blocklist-update.",
        ),
        "cache-size-mb": (
            "number",
            10,
            None,
            None,
            None,
            "The maximum size of the disk cache in MB",
        ),
        "dht-enabled": ("boolean", 6, None, None, None, "Enables DHT."),
        "download-dir": (
            "string",
            1,
            None,
            None,
            None,
            "Set the session download directory.",
        ),
        "download-queue-size": (
            "number",
            14,
            None,
            None,
            None,
            "Number of slots in the download queue.",
        ),
        "download-queue-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "Enables download queue.",
        ),
        "encryption": (
            "string",
            1,
            None,
            None,
            None,
            "Set the session encryption mode, one of ``required``, ``preferred`` or ``tolerated``.",
        ),
        "idle-seeding-limit": (
            "number",
            10,
            None,
            None,
            None,
            "The default seed inactivity limit in minutes.",
        ),
        "idle-seeding-limit-enabled": (
            "boolean",
            10,
            None,
            None,
            None,
            "Enables the default seed inactivity limit",
        ),
        "incomplete-dir": (
            "string",
            7,
            None,
            None,
            None,
            "The path to the directory of incomplete transfer data.",
        ),
        "incomplete-dir-enabled": (
            "boolean",
            7,
            None,
            None,
            None,
            "Enables the incomplete transfer data directory. Otherwise data for incomplete transfers are stored in the download target.",
        ),
        "lpd-enabled": (
            "boolean",
            9,
            None,
            None,
            None,
            "Enables local peer discovery for public torrents.",
        ),
        "peer-limit": (
            "number",
            1,
            5,
            None,
            "peer-limit-global",
            "Maximum number of peers.",
        ),
        "peer-limit-global": (
            "number",
            5,
            None,
            "peer-limit",
            None,
            "Maximum number of peers.",
        ),
        "peer-limit-per-torrent": (
            "number",
            5,
            None,
            None,
            None,
            "Maximum number of peers per transfer.",
        ),
        "pex-allowed": (
            "boolean",
            1,
            5,
            None,
            "pex-enabled",
            "Allowing PEX in public torrents.",
        ),
        "pex-enabled": (
            "boolean",
            5,
            None,
            "pex-allowed",
            None,
            "Allowing PEX in public torrents.",
        ),
        "port": ("number", 1, 5, None, "peer-port", "Peer port."),
        "peer-port": ("number", 5, None, "port", None, "Peer port."),
        "peer-port-random-on-start": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables randomized peer port on start of Transmission.",
        ),
        "port-forwarding-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "Enables port forwarding.",
        ),
        "rename-partial-files": (
            "boolean",
            8,
            None,
            None,
            None,
            'Appends ".part" to incomplete files',
        ),
        "queue-stalled-minutes": (
            "number",
            14,
            None,
            None,
            None,
            "Number of minutes of idle that marks a transfer as stalled.",
        ),
        "queue-stalled-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "Enable tracking of stalled transfers.",
        ),
        "script-torrent-done-enabled": (
            "boolean",
            9,
            None,
            None,
            None,
            'Whether or not to call the "done" script.',
        ),
        "script-torrent-done-filename": (
            "string",
            9,
            None,
            None,
            None,
            "Filename of the script to run when the transfer is done.",
        ),
        "seed-queue-size": (
            "number",
            14,
            None,
            None,
            None,
            "Number of slots in the upload queue.",
        ),
        "seed-queue-enabled": (
            "boolean",
            14,
            None,
            None,
            None,
            "Enables upload queue.",
        ),
        "seedRatioLimit": (
            "double",
            5,
            None,
            None,
            None,
            "Seed ratio limit. 1.0 means 1:1 download and upload ratio.",
        ),
        "seedRatioLimited": (
            "boolean",
            5,
            None,
            None,
            None,
            "Enables seed ration limit.",
        ),
        "speed-limit-down": (
            "number",
            1,
            None,
            None,
            None,
            "Download speed limit (in Kib/s).",
        ),
        "speed-limit-down-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "Enables download speed limiting.",
        ),
        "speed-limit-up": (
            "number",
            1,
            None,
            None,
            None,
            "Upload speed limit (in Kib/s).",
        ),
        "speed-limit-up-enabled": (
            "boolean",
            1,
            None,
            None,
            None,
            "Enables upload speed limiting.",
        ),
        "start-added-torrents": (
            "boolean",
            9,
            None,
            None,
            None,
            "Added torrents will be started right away.",
        ),
        "trash-original-torrent-files": (
            "boolean",
            9,
            None,
            None,
            None,
            "The .torrent file of added torrents will be deleted.",
        ),
        "utp-enabled": (
            "boolean",
            13,
            None,
            None,
            None,
            "Enables Micro Transport Protocol (UTP).",
        ),
    },
}



# -*- coding: utf-8 -*-
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.
# pylint: disable=line-too-long

import sys
import re
import time
import operator
import os
import base64
import json
import gzip
import unicodedata

from urllib.parse import urlparse
from urllib.request import urlopen
from urllib.error import HTTPError

from transmissionrpc.constants import DEFAULT_PORT, DEFAULT_TIMEOUT
from transmissionrpc.error import TransmissionError, HTTPHandlerError
from transmissionrpc.utils import (
    LOGGER,
    get_arguments,
    make_rpc_name,
    argument_value_convert,
    rpc_bool,
    is_logger_configured,
)
from transmissionrpc.httphandler import DefaultHTTPHandler
from transmissionrpc.torrent import Torrent
from transmissionrpc.session import Session


def debug_httperror(error):
    """
    Log the Transmission RPC HTTP error.
    """
    if sys.platform == "win32":
        msg = error.message.decode(sys.stdout.encoding)
    else:
        msg = error.message
    try:
        data = json.loads(error.data)
    except ValueError:
        data = error.data
    LOGGER.debug(
        json.dumps(
            {
                "response": {
                    "url": error.url,
                    "code": error.code,
                    "msg": msg,
                    "headers": error.headers,
                    "data": data,
                }
            },
            indent=2,
        )
    )


def parse_torrent_id(arg):
    """Parse an torrent id or torrent hashString."""
    torrent_id = None
    if isinstance(arg, int):
        # handle index
        torrent_id = int(arg)
    elif isinstance(arg, float):
        torrent_id = int(arg)
        if torrent_id != arg:
            torrent_id = None
    elif isinstance(arg, str):
        try:
            torrent_id = int(arg)
            if torrent_id >= 2 ** 31:
                torrent_id = None
        except (ValueError, TypeError):
            pass
        if torrent_id is None:
            # handle hashes
            try:
                int(arg, 16)
                torrent_id = arg
            except (ValueError, TypeError):
                pass
    return torrent_id


def parse_torrent_ids(args):
    """
    Take things and make them valid torrent identifiers
    """
    ids = []

    if args is None:
        pass
    elif isinstance(args, str):
        for item in re.split("[ ,]+", args):
            if not item:
                continue
            addition = None
            torrent_id = parse_torrent_id(item)
            if torrent_id is not None:
                addition = [torrent_id]
            if not addition:
                # handle index ranges i.e. 5:10
                match = re.match(r"^(\d+):(\d+)$", item)
                if match:
                    try:
                        idx_from = int(match.group(1))
                        idx_to = int(match.group(2))
                        addition = list(range(idx_from, idx_to + 1))
                    except ValueError:
                        pass
            if not addition:
                raise ValueError('Invalid torrent id, "%s"' % item)
            ids.extend(addition)
    elif isinstance(args, (list, tuple)):
        for item in args:
            ids.extend(parse_torrent_ids(item))
    else:
        torrent_id = parse_torrent_id(args)
        if torrent_id is None:
            raise ValueError("Invalid torrent id")
        else:
            ids = [torrent_id]
    return ids


class Client:
    """
    Client is the class handling the Transmission JSON-RPC client protocol.
    """

    def __init__(
        self,
        address="localhost",
        port=DEFAULT_PORT,
        user=None,
        password=None,
        http_handler=None,
        timeout=None,
    ):
        if isinstance(timeout, (int, float)):
            self._query_timeout = float(timeout)
        else:
            self._query_timeout = DEFAULT_TIMEOUT
        urlo = urlparse(address)
        if urlo.scheme == "":
            base_url = "http://" + address + ":" + str(port)
            self.url = base_url + "/transmission/rpc"
        else:
            if urlo.port:
                self.url = (
                    urlo.scheme
                    + "://"
                    + urlo.hostname
                    + ":"
                    + str(urlo.port)
                    + urlo.path
                )
            else:
                self.url = urlo.scheme + "://" + urlo.hostname + urlo.path
            LOGGER.info('Using custom URL "%s".', self.url)
            if urlo.username and urlo.password:
                user = urlo.username
                password = urlo.password
            elif urlo.username or urlo.password:
                LOGGER.warning(
                    "Either user or password missing, not using authentication."
                )
        if http_handler is None:
            self.http_handler = DefaultHTTPHandler()
        else:
            if hasattr(http_handler, "set_authentication") and hasattr(
                http_handler, "request"
            ):
                self.http_handler = http_handler
            else:
                raise ValueError("Invalid HTTP handler.")
        if user and password:
            self.http_handler.set_authentication(self.url, user, password)
        elif user or password:
            LOGGER.warning("Either user or password missing, not using authentication.")
        self._sequence = 0
        self.session = None
        self.session_id = 0
        self.server_version = None
        self.protocol_version = None
        self.get_session()
        self.torrent_get_arguments = get_arguments("torrent-get", self.rpc_version)

    def _get_timeout(self):
        """
        Get current timeout for HTTP queries.
        """
        return self._query_timeout

    def _set_timeout(self, value):
        """
        Set timeout for HTTP queries.
        """
        self._query_timeout = float(value)

    def _del_timeout(self):
        """
        Reset the HTTP query timeout to the default.
        """
        self._query_timeout = DEFAULT_TIMEOUT

    timeout = property(
        _get_timeout, _set_timeout, _del_timeout, doc="HTTP query timeout."
    )

    def _http_query(self, query, timeout=None):
        """
        Query Transmission through HTTP.
        """
        headers = {"x-transmission-session-id": str(self.session_id)}
        result = {}
        request_count = 0
        if timeout is None:
            timeout = self._query_timeout
        use_logger = is_logger_configured()
        while True:
            if use_logger:
                LOGGER.debug(
                    json.dumps(
                        {
                            "url": self.url,
                            "headers": headers,
                            "query": query,
                            "timeout": timeout,
                        },
                        indent=2,
                    )
                )
            try:
                result = self.http_handler.request(self.url, query, headers, timeout)
                break
            except HTTPHandlerError as error:
                if error.code == 409:
                    if use_logger:
                        LOGGER.info(
                            "Server responded with 409, trying to set session-id."
                        )
                    if request_count > 1:
                        raise TransmissionError("Session ID negotiation failed.", error)
                    session_id = None
                    for key in list(error.headers.keys()):
                        if key.lower() == "x-transmission-session-id":
                            session_id = error.headers[key]
                            self.session_id = session_id
                            headers = {
                                "x-transmission-session-id": str(self.session_id)
                            }
                    if session_id is None:
                        if use_logger:
                            debug_httperror(error)
                        raise TransmissionError("Unknown conflict.", error)
                else:
                    if use_logger:
                        debug_httperror(error)
                    raise TransmissionError("Request failed.", error)
            request_count += 1
        return result

    def _request(
        self, method, arguments=None, ids=None, require_ids=False, timeout=None
    ):
        """
        Send json-rpc request to Transmission using http POST
        """
        if not isinstance(method, str):
            raise ValueError("request takes method as string")
        if arguments is None:
            arguments = {}
        if not isinstance(arguments, dict):
            raise ValueError("request takes arguments as dict")
        ids = parse_torrent_ids(ids)
        if ids:
            arguments["ids"] = ids
        elif require_ids:
            raise ValueError("request require ids")
        use_logger = is_logger_configured()

        query = json.dumps(
            {"tag": self._sequence, "method": method, "arguments": arguments}
        )
        self._sequence += 1
        start = time.time()
        http_data = self._http_query(query, timeout)
        if isinstance(http_data, bytes):
            http_data = str(http_data, encoding="utf-8", errors="replace")
        http_data = "".join(
            list(filter(lambda c: unicodedata.category(c)[0] != "C", http_data))
        )
        elapsed = time.time() - start
        if use_logger:
            LOGGER.info("http request took %.3f s", elapsed)

        try:
            data = json.loads(http_data)
        except ValueError as error:
            if use_logger:
                LOGGER.error("Error: %s", str(error))
                LOGGER.error('Request: "%s"', query)
                LOGGER.error('HTTP data: "%s"', http_data)
            raise

        if use_logger:
            LOGGER.debug(json.dumps(data, indent=2))
        if "result" in data:
            if data["result"] != "success":
                raise TransmissionError(
                    'Query failed with result "%s".' % (data["result"])
                )
        else:
            raise TransmissionError("Query failed without result.")

        results = {}
        if method == "torrent-get":
            for item in data["arguments"]["torrents"]:
                results[item["id"]] = Torrent(self, item)
                if self.protocol_version == 2 and "peers" not in item:
                    self.protocol_version = 1
        elif method == "torrent-add":
            item = None
            if "torrent-added" in data["arguments"]:
                item = data["arguments"]["torrent-added"]
            elif "torrent-duplicate" in data["arguments"]:
                item = data["arguments"]["torrent-duplicate"]
            if item:
                results[item["id"]] = Torrent(self, item)
            else:
                raise TransmissionError("Invalid torrent-add response.")
        elif method == "session-get":
            self._update_session(data["arguments"])
        elif method == "session-stats":
            # older versions of T has the return data in "session-stats"
            if "session-stats" in data["arguments"]:
                self._update_session(data["arguments"]["session-stats"])
            else:
                self._update_session(data["arguments"])
        elif method in (
            "port-test",
            "blocklist-update",
            "free-space",
            "torrent-rename-path",
        ):
            results = data["arguments"]
        else:
            return None

        return results

    def _update_session(self, data):
        """
        Update session data.
        """
        if self.session:
            self.session.from_request(data)
        else:
            self.session = Session(self, data)

    def _update_server_version(self):
        """Decode the Transmission version string, if available."""
        if self.server_version is None:
            version_major = 1
            version_minor = 30
            version_changeset = 0
            version_parser = re.compile(r"(\d).(\d+) \((\d+)\)")
            if hasattr(self.session, "version"):
                match = version_parser.match(self.session.version)
                if match:
                    version_major = int(match.group(1))
                    version_minor = int(match.group(2))
                    version_changeset = match.group(3)
            self.server_version = (version_major, version_minor, version_changeset)

    @property
    def rpc_version(self):
        """
        Get the Transmission RPC version. Trying to deduct if the server don't have a version value.
        """
        if self.protocol_version is None:
            # Ugly fix for 2.20 - 2.22 reporting rpc-version 11, but having new arguments
            if self.server_version and (
                self.server_version[0] == 2 and self.server_version[1] in [20, 21, 22]
            ):
                self.protocol_version = 12
            # Ugly fix for 2.12 reporting rpc-version 10, but having new arguments
            elif self.server_version and (
                self.server_version[0] == 2 and self.server_version[1] == 12
            ):
                self.protocol_version = 11
            elif hasattr(self.session, "rpc_version"):
                self.protocol_version = self.session.rpc_version
            elif hasattr(self.session, "version"):
                self.protocol_version = 3
            else:
                self.protocol_version = 2
        return self.protocol_version

    def _rpc_version_warning(self, version):
        """
        Add a warning to the log if the Transmission RPC version is lower then the provided version.
        """
        if self.rpc_version < version:
            LOGGER.warning(
                "Using feature not supported by server. RPC version for server %d, feature introduced in %d.",
                self.rpc_version,
                version,
            )

    def add_torrent(self, torrent, timeout=None, **kwargs):
        """
        Add torrent to transfers list. Takes a uri to a torrent or base64 encoded torrent data in ``torrent``.
        Additional arguments are:

        ===================== ===== =========== =============================================================
        Argument              RPC   Replaced by Description
        ===================== ===== =========== =============================================================
        ``bandwidthPriority`` 8 -               Priority for this transfer.
        ``cookies``           13 -              One or more HTTP cookie(s).
        ``download_dir``      1 -               The directory where the downloaded contents will be saved in.
        ``files_unwanted``    1 -               A list of file id's that shouldn't be downloaded.
        ``files_wanted``      1 -               A list of file id's that should be downloaded.
        ``paused``            1 -               If True, does not start the transfer when added.
        ``peer_limit``        1 -               Maximum number of peers allowed.
        ``priority_high``     1 -               A list of file id's that should have high priority.
        ``priority_low``      1 -               A list of file id's that should have low priority.
        ``priority_normal``   1 -               A list of file id's that should have normal priority.
        ===================== ===== =========== =============================================================

        Returns a Torrent object with the fields.
        """
        if torrent is None:
            raise ValueError("add_torrent requires data or a URI.")
        torrent_data = None
        parsed_uri = urlparse(torrent)
        if parsed_uri.scheme in ["ftp", "ftps", "http", "https"]:
            # there has been some problem with T's built in torrent fetcher,
            # use a python one instead
            try:
                torrent_file = urlopen(torrent)
                if torrent_file.info().get("Content-Encoding") == "gzip":
                    torrent_data = gzip.decompress(torrent_file.read())
                else:
                    torrent_data = torrent_file.read()
                torrent_data = base64.b64encode(torrent_data).decode("utf-8")
            except HTTPError as e:
                # follow links that directly redirect to a magnet (redirects to
                # non-http/ftp files return a 302 error)
                if e.code == 302 and e.url.startswith("magnet:"):
                    return self.add_torrent(e.url, timeout=timeout, **kwargs)
                else:
                    raise e
        if parsed_uri.scheme in ["file"]:
            filepath = torrent
            # uri decoded different on linux / windows ?
            if parsed_uri.path:
                filepath = parsed_uri.path
            elif parsed_uri.netloc:
                filepath = parsed_uri.netloc
            if filepath.endswith(".gz"):
                with gzip.open(filepath, "rb") as torrent_file:
                    torrent_data = torrent_file.read()
            else:
                with open(filepath, "rb") as torrent_file:
                    torrent_data = torrent_file.read()
            torrent_data = base64.b64encode(torrent_data).decode("utf-8")
        if not torrent_data:
            if torrent.endswith(".torrent") or torrent.startswith("magnet:"):
                torrent_data = None
            else:
                might_be_base64 = False
                try:
                    base64.b64decode(torrent.encode("utf-8"))
                    might_be_base64 = True
                except Exception:
                    pass
                if might_be_base64:
                    torrent_data = torrent
        args = {}
        if torrent_data:
            args = {"metainfo": torrent_data}
        else:
            args = {"filename": torrent}
        for key, value in list(kwargs.items()):
            argument = make_rpc_name(key)
            (arg, val) = argument_value_convert(
                "torrent-add", argument, value, self.rpc_version
            )
            args[arg] = val
        return list(self._request("torrent-add", args, timeout=timeout).values())[0]

    def remove_torrent(self, ids, delete_data=False, timeout=None):
        """
        remove torrent(s) with provided id(s). Local data is removed if
        delete_data is True, otherwise not.
        """
        self._rpc_version_warning(3)
        self._request(
            "torrent-remove",
            {"delete-local-data": rpc_bool(delete_data)},
            ids,
            True,
            timeout=timeout,
        )

    def start_torrent(self, ids, bypass_queue=False, timeout=None):
        """Start torrent(s) with provided id(s)"""
        method = "torrent-start"
        if bypass_queue and self.rpc_version >= 14:
            method = "torrent-start-now"
        self._request(method, {}, ids, True, timeout=timeout)

    def start_all(self, bypass_queue=False, timeout=None):
        """Start all torrents respecting the queue order"""
        torrent_list = self.get_torrents()
        method = "torrent-start"
        if self.rpc_version >= 14:
            if bypass_queue:
                method = "torrent-start-now"
            torrent_list = sorted(
                torrent_list, key=operator.attrgetter("queuePosition")
            )
        ids = [x.id for x in torrent_list]
        self._request(method, {}, ids, True, timeout=timeout)

    def stop_torrent(self, ids, timeout=None):
        """stop torrent(s) with provided id(s)"""
        self._request("torrent-stop", {}, ids, True, timeout=timeout)

    def verify_torrent(self, ids, timeout=None):
        """verify torrent(s) with provided id(s)"""
        self._request("torrent-verify", {}, ids, True, timeout=timeout)

    def reannounce_torrent(self, ids, timeout=None):
        """Reannounce torrent(s) with provided id(s)"""
        self._rpc_version_warning(5)
        self._request("torrent-reannounce", {}, ids, True, timeout=timeout)

    def get_torrent(self, torrent_id, arguments=None, timeout=None):
        """
        Get information for torrent with provided id.
        ``arguments`` contains a list of field names to be returned, when None
        all fields are requested. See the Torrent class for more information.

        Returns a Torrent object with the requested fields.
        """
        if not arguments:
            arguments = self.torrent_get_arguments
        torrent_id = parse_torrent_id(torrent_id)
        if torrent_id is None:
            raise ValueError("Invalid id")
        result = self._request(
            "torrent-get",
            {"fields": arguments},
            torrent_id,
            require_ids=True,
            timeout=timeout,
        )
        if torrent_id in result:
            return result[torrent_id]
        else:
            for torrent in list(result.values()):
                if torrent.hashString == torrent_id:
                    return torrent
            raise KeyError("Torrent not found in result")

    def get_torrents(self, ids=None, arguments=None, timeout=None):
        """
        Get information for torrents with provided ids. For more information see get_torrent.

        Returns a list of Torrent object.
        """
        if not arguments:
            arguments = self.torrent_get_arguments
        return list(
            self._request(
                "torrent-get", {"fields": arguments}, ids, timeout=timeout
            ).values()
        )

    def get_files(self, ids=None, timeout=None):
        """
        Get list of files for provided torrent id(s). If ids is empty,
        information for all torrents are fetched. This function returns a dictionary
        for each requested torrent id holding the information about the files.

        ::

            {
                <torrent id>: {
                    <file id>: {
                        'name': <file name>,
                        'size': <file size in bytes>,
                        'completed': <bytes completed>,
                        'priority': <priority ('high'|'normal'|'low')>,
                        'selected': <selected for download (True|False)>
                    }

                    ...
                }

                ...
            }
        """
        fields = ["id", "name", "hashString", "files", "priorities", "wanted"]
        request_result = self._request(
            "torrent-get", {"fields": fields}, ids, timeout=timeout
        )
        result = {}
        for tid, torrent in list(request_result.items()):
            result[tid] = torrent.files()
        return result

    def set_files(self, items, timeout=None):
        """
        Set file properties. Takes a dictionary with similar contents as the result
        of `get_files`.

        ::

            {
                <torrent id>: {
                    <file id>: {
                        'priority': <priority ('high'|'normal'|'low')>,
                        'selected': <selected for download (True|False)>
                    }

                    ...
                }

                ...
            }
        """
        if not isinstance(items, dict):
            raise ValueError("Invalid file description")
        for tid, files in list(items.items()):
            if not isinstance(files, dict):
                continue
            wanted = []
            unwanted = []
            high = []
            normal = []
            low = []
            for fid, file_desc in list(files.items()):
                if not isinstance(file_desc, dict):
                    continue
                if "selected" in file_desc and file_desc["selected"]:
                    wanted.append(fid)
                else:
                    unwanted.append(fid)
                if "priority" in file_desc:
                    if file_desc["priority"] == "high":
                        high.append(fid)
                    elif file_desc["priority"] == "normal":
                        normal.append(fid)
                    elif file_desc["priority"] == "low":
                        low.append(fid)
            args = {"timeout": timeout}
            if high:
                args["priority_high"] = high
            if normal:
                args["priority_normal"] = normal
            if low:
                args["priority_low"] = low
            if wanted:
                args["files_wanted"] = wanted
            if unwanted:
                args["files_unwanted"] = unwanted
            self.change_torrent([tid], **args)

    def change_torrent(self, ids, timeout=None, **kwargs):
        """
        Change torrent parameters for the torrent(s) with the supplied id's. The
        parameters are:

        ============================ ===== =============== =======================================================================================
        Argument                     RPC   Replaced by     Description
        ============================ ===== =============== =======================================================================================
        ``bandwidthPriority``        5 -                   Priority for this transfer.
        ``downloadLimit``            5 -                   Set the speed limit for download in Kib/s.
        ``downloadLimited``          5 -                   Enable download speed limiter.
        ``files_unwanted``           1 -                   A list of file id's that shouldn't be downloaded.
        ``files_wanted``             1 -                   A list of file id's that should be downloaded.
        ``honorsSessionLimits``      5 -                   Enables or disables the transfer to honour the upload limit set in the session.
        ``location``                 1 -                   Local download location.
        ``peer_limit``               1 -                   The peer limit for the torrents.
        ``priority_high``            1 -                   A list of file id's that should have high priority.
        ``priority_low``             1 -                   A list of file id's that should have normal priority.
        ``priority_normal``          1 -                   A list of file id's that should have low priority.
        ``queuePosition``            14 -                  Position of this transfer in its queue.
        ``seedIdleLimit``            10 -                  Seed inactivity limit in minutes.
        ``seedIdleMode``             10 -                  Seed inactivity mode. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.
        ``seedRatioLimit``           5 -                   Seeding ratio.
        ``seedRatioMode``            5 -                   Which ratio to use. 0 = Use session limit, 1 = Use transfer limit, 2 = Disable limit.
        ``speed_limit_down``         1 - 5 downloadLimit   Set the speed limit for download in Kib/s.
        ``speed_limit_down_enabled`` 1 - 5 downloadLimited Enable download speed limiter.
        ``speed_limit_up``           1 - 5 uploadLimit     Set the speed limit for upload in Kib/s.
        ``speed_limit_up_enabled``   1 - 5 uploadLimited   Enable upload speed limiter.
        ``trackerAdd``               10 -                  Array of string with announce URLs to add.
        ``trackerRemove``            10 -                  Array of ids of trackers to remove.
        ``trackerReplace``           10 -                  Array of (id, url) tuples where the announce URL should be replaced.
        ``uploadLimit``              5 -                   Set the speed limit for upload in Kib/s.
        ``uploadLimited``            5 -                   Enable upload speed limiter.
        ============================ ===== =============== =======================================================================================

        .. NOTE::
           transmissionrpc will try to automatically fix argument errors.
        """
        args = {}
        for key, value in list(kwargs.items()):
            argument = make_rpc_name(key)
            (arg, val) = argument_value_convert(
                "torrent-set", argument, value, self.rpc_version
            )
            args[arg] = val

        if args:
            self._request("torrent-set", args, ids, True, timeout=timeout)
        else:
            ValueError("No arguments to set")

    def move_torrent_data(self, ids, location, timeout=None):
        """Move torrent data to the new location."""
        self._rpc_version_warning(6)
        args = {"location": location, "move": True}
        self._request("torrent-set-location", args, ids, True, timeout=timeout)

    def locate_torrent_data(self, ids, location, timeout=None):
        """Locate torrent data at the provided location."""
        self._rpc_version_warning(6)
        args = {"location": location, "move": False}
        self._request("torrent-set-location", args, ids, True, timeout=timeout)

    def rename_torrent_path(self, torrent_id, location, name, timeout=None):
        """
        Rename directory and/or files for torrent.
        Remember to use get_torrent or get_torrents to update your file information.
        """
        self._rpc_version_warning(15)
        torrent_id = parse_torrent_id(torrent_id)
        if torrent_id is None:
            raise ValueError("Invalid id")
        dirname = os.path.dirname(name)
        if dirname:
            raise ValueError("Target name cannot contain a path delimiter")
        args = {"path": location, "name": name}
        result = self._request(
            "torrent-rename-path", args, torrent_id, True, timeout=timeout
        )
        return (result["path"], result["name"])

    def queue_top(self, ids, timeout=None):
        """Move transfer to the top of the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-top", ids=ids, require_ids=True, timeout=timeout)

    def queue_bottom(self, ids, timeout=None):
        """Move transfer to the bottom of the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-bottom", ids=ids, require_ids=True, timeout=timeout)

    def queue_up(self, ids, timeout=None):
        """Move transfer up in the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-up", ids=ids, require_ids=True, timeout=timeout)

    def queue_down(self, ids, timeout=None):
        """Move transfer down in the queue."""
        self._rpc_version_warning(14)
        self._request("queue-move-down", ids=ids, require_ids=True, timeout=timeout)

    def get_session(self, timeout=None):
        """
        Get session parameters. See the Session class for more information.
        """
        if self.session is None:
            self._request("session-get", timeout=timeout)
            self._update_server_version()
        return self.session

    def set_session(self, timeout=None, **kwargs):
        """
        Set session parameters. The parameters are:

        ================================ ===== ================= ==========================================================================================================================
        Argument                         RPC   Replaced by       Description
        ================================ ===== ================= ==========================================================================================================================
        ``alt_speed_down``               5 -                     Alternate session download speed limit (in Kib/s).
        ``alt_speed_enabled``            5 -                     Enables alternate global download speed limiter.
        ``alt_speed_time_begin``         5 -                     Time when alternate speeds should be enabled. Minutes after midnight.
        ``alt_speed_time_day``           5 -                     Enables alternate speeds scheduling these days.
        ``alt_speed_time_enabled``       5 -                     Enables alternate speeds scheduling.
        ``alt_speed_time_end``           5 -                     Time when alternate speeds should be disabled. Minutes after midnight.
        ``alt_speed_up``                 5 -                     Alternate session upload speed limit (in Kib/s).
        ``blocklist_enabled``            5 -                     Enables the block list
        ``blocklist_url``                11 -                    Location of the block list. Updated with blocklist-update.
        ``cache_size_mb``                10 -                    The maximum size of the disk cache in MB
        ``dht_enabled``                  6 -                     Enables DHT.
        ``download_dir``                 1 -                     Set the session download directory.
        ``download_queue_enabled``       14 -                    Enables download queue.
        ``download_queue_size``          14 -                    Number of slots in the download queue.
        ``encryption``                   1 -                     Set the session encryption mode, one of ``required``, ``preferred`` or ``tolerated``.
        ``idle_seeding_limit``           10 -                    The default seed inactivity limit in minutes.
        ``idle_seeding_limit_enabled``   10 -                    Enables the default seed inactivity limit
        ``incomplete_dir``               7 -                     The path to the directory of incomplete transfer data.
        ``incomplete_dir_enabled``       7 -                     Enables the incomplete transfer data directory. Otherwise data for incomplete transfers are stored in the download target.
        ``lpd_enabled``                  9 -                     Enables local peer discovery for public torrents.
        ``peer_limit``                   1 - 5 peer-limit-global Maximum number of peers.
        ``peer_limit_global``            5 -                     Maximum number of peers.
        ``peer_limit_per_torrent``       5 -                     Maximum number of peers per transfer.
        ``peer_port``                    5 -                     Peer port.
        ``peer_port_random_on_start``    5 -                     Enables randomized peer port on start of Transmission.
        ``pex_allowed``                  1 - 5 pex-enabled       Allowing PEX in public torrents.
        ``pex_enabled``                  5 -                     Allowing PEX in public torrents.
        ``port``                         1 - 5 peer-port         Peer port.
        ``port_forwarding_enabled``      1 -                     Enables port forwarding.
        ``queue_stalled_enabled``        14 -                    Enable tracking of stalled transfers.
        ``queue_stalled_minutes``        14 -                    Number of minutes of idle that marks a transfer as stalled.
        ``rename_partial_files``         8 -                     Appends ".part" to incomplete files
        ``script_torrent_done_enabled``  9 -                     Whether or not to call the "done" script.
        ``script_torrent_done_filename`` 9 -                     Filename of the script to run when the transfer is done.
        ``seed_queue_enabled``           14 -                    Enables upload queue.
        ``seed_queue_size``              14 -                    Number of slots in the upload queue.
        ``seedRatioLimit``               5 -                     Seed ratio limit. 1.0 means 1:1 download and upload ratio.
        ``seedRatioLimited``             5 -                     Enables seed ration limit.
        ``speed_limit_down``             1 -                     Download speed limit (in Kib/s).
        ``speed_limit_down_enabled``     1 -                     Enables download speed limiting.
        ``speed_limit_up``               1 -                     Upload speed limit (in Kib/s).
        ``speed_limit_up_enabled``       1 -                     Enables upload speed limiting.
        ``start_added_torrents``         9 -                     Added torrents will be started right away.
        ``trash_original_torrent_files`` 9 -                     The .torrent file of added torrents will be deleted.
        ``utp_enabled``                  13 -                    Enables Micro Transport Protocol (UTP).
        ================================ ===== ================= ==========================================================================================================================

        .. NOTE::
           transmissionrpc will try to automatically fix argument errors.
        """
        args = {}
        for key, value in list(kwargs.items()):
            if key == "encryption" and value not in [
                "required",
                "preferred",
                "tolerated",
            ]:
                raise ValueError("Invalid encryption value")
            argument = make_rpc_name(key)
            (arg, val) = argument_value_convert(
                "session-set", argument, value, self.rpc_version
            )
            args[arg] = val
        if args:
            self._request("session-set", args, timeout=timeout)

    def blocklist_update(self, timeout=None):
        """Update block list. Returns the size of the block list."""
        self._rpc_version_warning(5)
        result = self._request("blocklist-update", timeout=timeout)
        if "blocklist-size" in result:
            return result["blocklist-size"]
        return None

    def port_test(self, timeout=None):
        """
        Tests to see if your incoming peer port is accessible from the
        outside world.
        """
        self._rpc_version_warning(5)
        result = self._request("port-test", timeout=timeout)
        if "port-is-open" in result:
            return result["port-is-open"]
        return None

    def free_space(self, path, timeout=None):
        """
        Get the ammount of free space (in bytes) at the provided location.
        """
        self._rpc_version_warning(15)
        result = self._request("free-space", {"path": path}, timeout=timeout)
        if result["path"] == path:
            return result["size-bytes"]
        return None

    def session_stats(self, timeout=None):
        """Get session statistics"""
        self._request("session-stats", timeout=timeout)
        return self.session



# -*- coding: utf-8 -*-
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.


class TransmissionError(Exception):
    """
    This exception is raised when there has occurred an error related to
    communication with Transmission. It is a subclass of Exception.
    """

    def __init__(self, message="", original=None):
        Exception.__init__(self)
        self.message = message
        self.original = original

    def __str__(self):
        if self.original:
            original_name = type(self.original).__name__
            return '%s Original exception: %s, "%s"' % (
                self.message,
                original_name,
                str(self.original),
            )
        else:
            return self.message


class HTTPHandlerError(Exception):
    """
    This exception is raised when there has occurred an error related to
    the HTTP handler. It is a subclass of Exception.
    """

    def __init__(
        self, httpurl=None, httpcode=None, httpmsg=None, httpheaders=None, httpdata=None
    ):
        Exception.__init__(self)
        self.url = ""
        self.code = 600
        self.message = ""
        self.headers = {}
        self.data = ""
        if isinstance(httpurl, str):
            self.url = httpurl
        if isinstance(httpcode, int):
            self.code = httpcode
        if isinstance(httpmsg, str):
            self.message = httpmsg
        if isinstance(httpheaders, dict):
            self.headers = httpheaders
        if isinstance(httpdata, str):
            self.data = httpdata

    def __repr__(self):
        return "<HTTPHandlerError %d, %s>" % (self.code, self.message)

    def __str__(self):
        return "HTTPHandlerError %d: %s" % (self.code, self.message)

    def __unicode__(self):
        return "HTTPHandlerError %d: %s" % (self.code, self.message)



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.

import re
from base64 import b64encode
from io import BytesIO

import pycurl
from transmissionrpc.error import HTTPHandlerError


class HTTPHandler:
    """
    Prototype for HTTP handling.
    """

    def set_authentication(self, uri, login, password):
        """
        Transmission use basic authentication in earlier versions and digest
        authentication in later versions.

         * uri, the authentication realm URI.
         * login, the authentication login.
         * password, the authentication password.
        """
        raise NotImplementedError(
            "Bad HTTPHandler, failed to implement set_authentication."
        )

    def request(self, url, query, headers, timeout):
        """
        Implement a HTTP POST request here.

         * url, The URL to request.
         * query, The query data to send. This is a JSON data string.
         * headers, a dictionary of headers to send.
         * timeout, requested request timeout in seconds.
        """
        raise NotImplementedError("Bad HTTPHandler, failed to implement request.")


class DefaultHTTPHandler(HTTPHandler):
    """
    The default HTTP handler provided with transmissionrpc.
    """

    def __init__(self):
        HTTPHandler.__init__(self)
        self._authorization = None

    def _response_encoding(self, headers):
        encoding = None

        # Figure out what encoding was sent with the response, if any.
        # Check against lowercased header name.
        if headers and "content-type" in headers:
            content_type = headers["content-type"].lower()
            match = re.search(r"charset=(\S+)", content_type)
            if match:
                encoding = match.group(1)

        if encoding is None:
            # Default encoding for HTML is iso-8859-1.
            # Other content types may have different default encoding,
            # or in case of binary data, may have no encoding at all.
            encoding = "iso-8859-1"

        return encoding

    def _response_headers(self, buffer):
        headers = {}

        # HTTP standard specifies that headers are encoded in iso-8859-1.
        for line in buffer.getvalue().decode("iso-8859-1").splitlines():
            # Header lines include the first status line (HTTP/1.x ...).
            # We are going to ignore all lines that don't have a colon in them.
            # This will botch headers that are split on multiple lines...
            if ":" not in line:
                continue

            # Break the header line into header name and value.
            name, value = line.split(":", 1)

            # Remove whitespace that may be present.
            # Header lines include the trailing newline, and there may be
            # whitespace around the colon.
            name = name.strip()
            value = value.strip()

            # Header names are case insensitive.
            # Lowercase name here.
            name = name.lower()

            # Now we can actually record the header name and value.
            # Note: this only works when headers are not duplicated.
            headers[name] = value

        return headers

    def set_authentication(self, uri, login, password):
        self._authorization = b64encode(
            bytearray("{}:{}".format(login, password), "utf-8")
        ).decode("utf-8")

    def request(self, url, query, headers, timeout):
        request_headers = ["{}: {}".format(k, v) for k, v in headers.items()]

        if self._authorization:
            request_headers += ["Authorization: Basic {}".format(self._authorization)]

        buf1, buf2 = BytesIO(), BytesIO()
        c = pycurl.Curl()
        c.setopt(pycurl.URL, url)
        c.setopt(pycurl.TIMEOUT, int(timeout))
        c.setopt(pycurl.POSTFIELDS, query)
        c.setopt(pycurl.HTTPHEADER, request_headers)
        c.setopt(pycurl.WRITEFUNCTION, buf1.write)
        c.setopt(pycurl.HEADERFUNCTION, buf2.write)
        c.perform()

        response_code = c.getinfo(pycurl.RESPONSE_CODE)
        c.close()
        response_headers = self._response_headers(buf2)
        encoding = self._response_encoding(response_headers)
        response_body = buf1.getvalue().decode(encoding)

        if response_code == 200:
            return response_body
        else:
            raise HTTPHandlerError(
                url, response_code, None, response_headers, response_body
            )



# -*- coding: utf-8 -*-
# Copyright (c) 2008-2014 Erik Svensson <erik.public@gmail.com>
# Copyright (c) 2019 Janne K <0x022b@gmail.com>
# Licensed under the MIT license.
# flake8: noqa

from transmissionrpc.constants import (
    DEFAULT_PORT,
    DEFAULT_TIMEOUT,
    PRIORITY,
    RATIO_LIMIT,
    LOGGER,
)
from transmissionrpc.error import TransmissionError, HTTPHandlerError
from transmissionrpc.httphandler import HTTPHandler, DefaultHTTPHandler
from transmissionrpc.torrent import Torrent
from transmissionrpc.session import Session
from transmissionrpc.client import Client
from transmissionrpc.utils import add_stdout_logger, add_file_logger

__author__ = "Janne K <0x022b@gmail.com>"
__version_major__ = 0
__version_minor__ = 14
__version_patch__ = 0
__version__ = "{0}.{1}.{2}".format(
    __version_major__, __version_minor__, __version_patch__
)
__copyright__ = "Copyright (c) 2008-2014 Erik Svensson, Copyright (c) 2019 Janne K"
__license__ = "MIT"
